#!/usr/bin/perl

######################################################################
## @brief   convert Blat pslx format to segemehl-like SAM
## @note    no SAM header is generated;
##          MD tag is not set (use samtools calmd);
##          paired-end data may not be properly converted
## @author  Christian Otto
## @date    Tue Aug 13 14:52:26 CEST 2013
## @version 0.1
######################################################################

die ("Usage: cat in.psl | convBl2Se.pl in.[fq|fa][.gz]\n")
    if (@ARGV == 0 || -t STDIN);

## global variables:
## - softclipping: clipped sequence is
##   present in SEQ
$SOFTCLIPPING = 1;

## read input fasta or fastq file
$IN = $ARGV[0];
if ($IN =~ m/\.gz$/){
    open(IN, "gunzip -c $IN |") or die;
}
else {
    open(IN, "<$IN") or die;
}
while(<IN>){
    chomp;
    if ($. % 2 == 1){
	$ch = substr($_, 0, 1, "");
	if ($ch ne "+"){
	    $id = $_;
	}
    }
    else {
	if ($ch ne "+"){
	    $SEQ{$id} = $_;
	}
	else {
	    $QUAL{$id} = $_;
	}
    }
}
close(IN);

## iterate through psl file
while($line = <STDIN>){
    chomp $line; 
    
    next unless $line =~ m/^\d/;
    
    @F = split("\t", $line);
    push(@{$READS{$F[9]}}, []);
    
    ## recalculate qry coordinates
    ## in case of minus strand
    if ($F[8] eq "-"){
	$tmp = $F[11];
	$F[11] = $F[10] - $F[12];
	$F[12] = $F[10] - $tmp;
    }

    $cigar = "";
    ## 5'-end clipping
    $cigar .= $F[11].($SOFTCLIPPING ? 'S' : 'H') if $F[11];

    ## split blocks
    @len = split(",", $F[18]);
    @qrypos = split(",", $F[19]);
    @refpos = split(",", $F[20]);

    ## process blocks
    $qrypos = $qrypos[0];
    $refpos = $refpos[0];
    for ($i = 0; $i < $F[17] - 1; $i++){
	$qryskip = $qrypos[$i+1]-$qrypos[$i]-$len[$i];
	$refskip = $refpos[$i+1]-$refpos[$i]-$len[$i];
	$cigar .= $len[$i].'M';

	## case 1: only skip at reference
	## if short  => deletion
	## otherwise => intron
	if ($qryskip == 0 && $refskip > 0){
	    if ($refskip > 10){
		$qryst = $qrypos + 1;
		$qryend = $qrypos[$i] + 1 + $len[$i] - 1;
		if ($F[8] eq '-'){
		    $tmp = $qryst;
		    $qryst = $F[10] - $qryend + 1;
		    $qryend = $F[10] - $tmp + 1;
		}
		$refst = $refpos + 1;
		$refend = $refpos[$i] + 1 + $len[$i] - 1;
		
		#print STDERR join("\t", ($qryst, $qryend, $qryskip, $refst, $refend, $refskip, $len[$i]))."\n";
		push(@{$READS{$F[9]}->[-1]}, [$F[9], (($F[8] eq '+')? 0 : 16), $F[13], $refst, 255, $cigar, 
					      '*', 0, 0, '*', '*', $qryst, $qryend, $refend]);
		$cigar = "";
		$refpos = $refpos[$i+1];
		$qrypos = $qrypos[$i+1];		
	    }
	    else {
		$cigar .= $refskip.'D';
	    }
	}
	## case 2: only skip at query
	## => insertion
	elsif ($refskip == 0 && $qryskip > 0){
	    $cigar .= $qryskip.'I';
	}
	## case 3: skip at reference & query
	## => split up (otherwise, realignment
	## would be necessary)
	else {
	    $qryst = $qrypos + 1;
	    $qryend = $qrypos[$i] + 1 + $len[$i] - 1;
	    if ($F[8] eq '-'){
		$tmp = $qryst;
		$qryst = $F[10] - $qryend + 1;
		$qryend = $F[10] - $tmp + 1;
	    }
	    $refst = $refpos + 1;
	    $refend = $refpos[$i] + 1 + $len[$i] - 1;

	    #print STDERR join("\t", ($qryst, $qryend, $qryskip, $refst, $refend, $refskip, $len[$i]))."\n";
	    push(@{$READS{$F[9]}->[-1]}, [$F[9], (($F[8] eq '+')? 0 : 16), $F[13], $refst, 255, $cigar, 
					  '*', 0, 0, '*', '*', $qryst, $qryend, $refend]);
	    $cigar = "";
	    $refpos = $refpos[$i+1];
	    $qrypos = $qrypos[$i+1];
	}
    }
    ## 3'-end clipping
    $cigar .= ($F[12] - $qrypos[$F[17] - 1]).'M';
    $cigar .= ($F[10] - $F[12]).($SOFTCLIPPING ? 'S' : 'H') if $F[10] != $F[12];

    $qryst = $qrypos + 1;
    $qryend = $qrypos[-1] + 1 + $len[-1] - 1;
    if ($F[8] eq '-'){
	$tmp = $qryst;
	$qryst = $F[10] - $qryend + 1;
	$qryend = $F[10] - $tmp + 1;
    }
    $refst = $refpos + 1;
    $refend = $refpos[-1] + 1 + $len[-1] - 1;

    #print STDERR join("\t", ($qryst, $qryend, $qryskip, $refst, $refend, $refskip, $len[$i]))."\n";
    push(@{$READS{$F[9]}->[-1]}, [$F[9], (($F[8] eq '+')? 0 : 16), $F[13], $refst, 255, $cigar, 
				  '*', 0, 0, '*', '*', $qryst, $qryend, $refend]);
}

foreach(sort(keys(%READS))){
    for ($k = 0; $k < scalar(@{$READS{$_}}); $k++){
	@frags = sort { $a->[-3] <=> $b->[-3] } @{$READS{$_}->[$k]};

	for ($i = 0; $i < scalar(@frags); $i++){
	    #print join("\t", @{$frags[$i]})."\n";
	    for ($j = $i + 1; $j < scalar(@frags); $j++){
		if (($frags[$i]->[-3] <= $frags[$j]->[-3] &&
		     $frags[$i]->[-2] >= $frags[$j]->[-3]) ||
		    ($frags[$i]->[-3] > $frags[$j]->[-3] &&
		     $frags[$i]->[-3] <= $frags[$j]->[-2])){
		    die "Error: overlapping query positions in $_\n";		
		}
	    }
	}

	for ($i=0; $i < scalar(@frags); $i++){
	    @res = @{$frags[$i]}[0..10];

	    ## add seq & qual
	    if ($SEQ{$res[0]}){
		$res[9] = "";
		if ($SOFTCLIPPING && $i == 0){
		    $res[9] .= substr($SEQ{$res[0]}, 0, $frags[$i]->[-3]-1);
		}
		$res[9] .= substr($SEQ{$res[0]}, $frags[$i]->[-3]-1, $frags[$i]->[-2] - $frags[$i]->[-3] + 1);
		if ($SOFTCLIPPING && $i == scalar(@frags) - 1){
		    $res[9] .= substr($SEQ{$res[0]}, $frags[$i]->[-2]);
		}
		if ($res[1] & 0x10){
		    $res[9] = reverse($res[9]);
		    $res[9] =~ tr/[ACGTNacgtn]/[TGCANtgcan]/;
		}
	    }
	    if ($QUAL{$res[0]}){
		$res[10] = "";		
		if ($SOFTCLIPPING && $i == 0){
		    $res[10] .= substr($QUAL{$res[0]}, 0, $frags[$i]->[-3]-1);
		}
		$res[10] .= substr($QUAL{$res[0]}, $frags[$i]->[-3]-1, $frags[$i]->[-2] - $frags[$i]->[-3] + 1);
		if ($SOFTCLIPPING && $i == scalar(@frags) - 1){
		    $res[10] .= substr($QUAL{$res[0]}, $frags[$i]->[-2]);
		}
		if ($res[1] & 0x10){
		    $res[10] = reverse($res[10]);
		}		
	    }
	    
	    ## add NH, XI
	    push(@res, "NH:i:".scalar(@{$READS{$_}}));
	    push(@res, "XI:i:".$k);
		 
	    ## add XL, XX, XY, and XQ
	    if (scalar(@frags) > 1){
		push(@res, "XL:i:".scalar(@frags));
		push(@res, "XX:i:".$frags[$i]->[-3]);
		push(@res, "XY:i:".$frags[$i]->[-2]);
		push(@res, "XQ:i:".$i);
		
		## set remaining segemehl tags
		if ($i != 0){
		    push(@res, "XP:Z:".$frags[$i-1]->[2]);
		    push(@res, "XU:i:".(($frags[$i-1]->[1] & 0x10)?
					$frags[$i-1]->[3] : $frags[$i-1]->[-1]));
		    push(@res, "XS:i:".(($frags[$i-1]->[1] & 0x10) ? 0 : 64));
		}
		if ($i != scalar(@frags) - 1){
		    push(@res, "XC:Z:".$frags[$i+1]->[2]);
		    push(@res, "XV:i:".(($frags[$i+1]->[1] & 0x10) ? 
					$frags[$i+1]->[-1] : $frags[$i+1]->[3]));
		    push(@res, "XT:i:".(($frags[$i+1]->[1] & 0x10) ? 0 : 32));
		}
	    }
	    print join("\t", @res)."\n";
	}
    }
}
