#!/usr/bin/perl

######################################################################
## @brief   convert TopHat format to segemehl-like SAM
## @note    NM and MD tags are not set (use samtools calmd);
##          paired-end data may not be properly converted
## @author  Christian Otto
## @date    Fri Jul  5 15:11:22 CEST 2013
## @version 0.1
######################################################################

while($line = <STDIN>){
    chomp $line; 
    
    do {print $line."\n"; next} if $line =~ /^@/;
    
    @F = split("\t", $line); 
    next if $F[1] & 0x4;
    die "Paired-end data not supported"
	if ($F[1] & 0x1) | ($F[1] & 0x40) | ($F[1] & 0x80) | $line =~ /XP:Z:/;

    ## set some initial variables
    $qrypos = 0;
    $qrylen = length($F[9]);
    $NH = $1 if $line =~ /NH:i:(\d+)/ or die;
    if ($NH > 1){
	$HI = $1 if $line =~ /HI:i:(\d+)/ or die;
    }
    else {
	$HI = 0;
    }

    ## in case of fusion alignments:
    ## - read & calculate $XF ($XF > 0 => fusion),
    ##   $XF_pos, $XF_cigar, $XF_off, and $XF_seq
    $XF = 0;
    foreach(@F){	
	if (/^XF:Z:/){
	    $XF = $1 if $_ =~ /XF:Z:(\d+)/;	    
	    (undef, undef, $XF_pos, $XF_cigar, $XF_seq) = split(/ /, $_);
	    ## XF_pos:
	    ## XF == 1 => use left-most query position
	    ## XF == 2 => use fusion position
	    if ($XF == 2){
		$XF_pos = $1 if $XF_cigar =~ /(\d+)F/ or die;
	    }

	    ## query offset:
	    ## - only set for 'second' fusion alignment
	    ##   which is $XF == 1 with alignments on minus
	    ##   and $XF == 2 with alignments on plus
	    ## - calculate offset using the mapping part
	    ##   of $XF_cigar before ($XF == 1) or after
	    ##   ($XF == 2) the fusion edge
	    $XF_off = 0;
	    if ($XF == 1 && ($F[1] & 0x10)){
		$tmp = $XF_cigar;
		$tmp =~ s/^.*\d+F//;
		$tmp =~ s/(\d+)[mx=iMX=I]/$XF_off+=$1/eg;
	    }
	    elsif ($XF == 2 && !($F[1] & 0x10)){
		$tmp = $XF_cigar;
		$tmp =~ s/\d+F.*$//;
		$tmp =~ s/(\d+)[mx=iMX=I]/$XF_off+=$1/eg;
	    }
	    last;
	}	
    }
    die if $XF < 0 || $XF > 2;

    if ($XF){
	## find stored alignment with same HI
	## or create new one
	for ($i = 0; $i < scalar(@{$READS{$F[0]}}); $i++){
	    last if ($READS{$F[0]}->[$i]->[0]->[11] == $HI);
	}
	if ($i == scalar(@{$READS{$F[0]}})){
	    push(@{$READS{$F[0]}}, []);
	}

	## correct mapping strand:
	## XF_pos != POS => swap strand
	if ($XF_pos != $F[3]){
	    $F[1] ^= 0x10;
	}
    }
    ## otherwise:
    ## create new alignment
    else {
	push(@{$READS{$F[0]}}, []);
	$i = scalar(@{$READS{$F[0]}})-1;
    }

    ## split up fragments by N in CIGAR
    while($F[5] =~ m/\d+N/){

	$F[5] =~ s/^([^N]+[MIDX=])(\d+)N//;
	$curcigar = $1; 
	$curjump = $2; 

	$qryst = $qrypos + 1;
	$tmp = $curcigar; 
	$clip5 = ($tmp =~ s/^(\d+)S//) ? $1 : 0;

	$qrypos += $clip5;
	$refst=$F[3];

	$tmp = $curcigar;
	$tmp =~ s/(\d+)[MX=D]/$F[3]+=$1/eg;
	$tmp = $curcigar;
	$tmp =~ s/(\d+)[MX=I]/$qrypos+=$1/eg;

	$qryend = $qrypos;
	$refend = $F[3]-1;
	$F[3] += $curjump;

	## qryst is before soft-clipping
	## qryend is after soft-clipping + matching
	## => current qrylen = $qryend-$qryst+1
	$curseq = substr($F[9], 0, $qryend-$qryst+1, "");
	if ($F[10] ne "*"){
	    $curqual = substr($F[10], 0, $qryend-$qryst+1, "");
	}
	else {
	    $curqual = "*";
	}

	if ($F[1] & 0x10){
	    $tmp = $qryst;
	    $qryst = $qrylen - $qrypos + 1;
	    $qryend = $qrylen - $tmp + 1;
	}

	## apply offset in case of fusion
	$qryst += $XF_off;
	$qryend += $XF_off;

	#print STDERR $XF_off."\t".$qryst."\t".$qryend."\t".$qrylen."\n";
	push(@{$READS{$F[0]}->[$i]}, [$F[0], $F[1], $F[2], $refst, $F[4], $curcigar, $F[6], $F[7], $F[8], 
				      $curseq, $curqual, $HI, $qryst, $qryend, $qrylen, $refend]);
    } 
    ## process last fragment
    $tmp = $F[5]; 
    $clip5 = ($tmp =~ s/^(\d+)S//) ? $1 : 0;
    
    $qrypos += $clip5;
    $qryst=$qrypos+1;

    $refst=$F[3];
    $refend = $F[3]-1;
    
    $tmp = $F[5];
    $tmp =~ s/(\d+)[MX=D]/$refend+=$1/eg;
    $tmp = $F[5];
    $tmp =~ s/(\d+)[MX=I]/$qrypos+=$1/eg;

    $qryend = $qrypos;
    if ($F[1] & 0x10){
	$tmp = $qryst;
	$qryst = $qrylen - $qrypos + 1;
	$qryend = $qrylen - $tmp + 1;
    }

    ## apply offset in case of fusion
    $qryst += $XF_off;
    $qryend += $XF_off;
    #print STDERR $XF_off."\t".$qryst."\t".$qryend."\t".$qrylen."\n";
    push(@{$READS{$F[0]}->[$i]}, [@F[0..10], $HI, $qryst, $qryend, $qrylen, $refend]);
}

foreach(sort(keys(%READS))){
    for ($k = 0; $k < scalar(@{$READS{$_}}); $k++){
	@frags = sort { $a->[-4] <=> $b->[-4] } @{$READS{$_}->[$k]};

	for ($i = 0; $i < scalar(@frags); $i++){
	    #print join("\t", @{$frags[$i]})."\n";
	    for ($j = $i + 1; $j < scalar(@frags); $j++){
		if (($frags[$i]->[-4] <= $frags[$j]->[-4] &&
		     $frags[$i]->[-3] >= $frags[$j]->[-4]) ||
		    ($frags[$i]->[-4] > $frags[$j]->[-4] &&
		     $frags[$i]->[-4] <= $frags[$j]->[-3])){
		    die "Error: overlapping query positions in $_\n";		
		}
	    }
	}
	
	for ($i=0; $i < scalar(@frags); $i++){
	    @res = @{$frags[$i]}[0..10];

	    ## add NH, XI
	    push(@res, "NH:i:".scalar(@{$READS{$_}}));
	    push(@res, "XI:i:".$k);
		 
	    ## add XL, XX, XY, and XQ
	    if (scalar(@frags) > 1){
		push(@res, "XL:i:".scalar(@frags));
		push(@res, "XX:i:".$frags[$i]->[-4]);
		push(@res, "XY:i:".$frags[$i]->[-3]);
		push(@res, "XQ:i:".$i);
		
		## set remaining segemehl tags (see convTo2Se.pl)
		if ($i != 0){
		    push(@res, "XP:Z:".$frags[$i-1]->[2]);
		    push(@res, "XU:i:".(($frags[$i-1]->[1] & 0x10)?
					$frags[$i-1]->[3] : $frags[$i-1]->[-1]));
		    push(@res, "XS:i:".(($frags[$i-1]->[1] & 0x10) ? 0 : 64));
		}
		if ($i != scalar(@frags) - 1){
		    push(@res, "XC:Z:".$frags[$i+1]->[2]);
		    push(@res, "XV:i:".(($frags[$i+1]->[1] & 0x10) ? 
					$frags[$i+1]->[-1] : $frags[$i+1]->[3]));
		    push(@res, "XT:i:".(($frags[$i+1]->[1] & 0x10) ? 0 : 32));
		}
	    }
	    print join("\t", @res)."\n";
	}
    }
}

